// COSC 4P76 - Project
// Michael Gircys (mg12vp, 5269550)
// 2016-04-21

package ec.app.COSC4P76_Proj;

// Interactive Evolutionary Art System

import ec.EvolutionState;
import ec.Individual;
import ec.gp.koza.KozaFitness;
import ec.util.Parameter;

import java.awt.image.BufferedImage;
import java.util.ArrayList;
import java.util.List;

import weka.core.Instance;
import weka.core.Instances;
import ec.app.IEAS.IEAS;
import ec.app.IEAS.MultiData;

// An extension of the base IEAS which uses an adaptive aesthetic learning system
// to grant the user the ability to defer image rating to the system.
@SuppressWarnings("serial")
public class IEASLearning extends IEAS implements IEASLearningProblemForm
{
	
	/* Inherited:
	 * The current position for which to evaluate the trees.
	public double Current_Pos_X;
	public double Current_Pos_Y;
	*/
	
	LearningSystem Learner;
	IEASSimpleAutomaticFitness AutoFitness = null;
	
	// Setup the IEAS learning problem
	/* Params (from ecj .params file):
	 * 		learning: base path from which further learning system parameters are specified
	 *		autofitness: if specified, the class providing the fixed fitness scheme 
	 */
	@Override
	public void setup(final EvolutionState state, final Parameter base)
    {
        super.setup(state, base);
        
        // verify our input is the right class (or subclasses from it)
        if (!(input instanceof MultiData))
            state.output.fatal("GPData class must subclass from " + MultiData.class, base.push(P_DATA), null);
        
        // Initialize learning systems
        Parameter p_learn = base.push("learning");
        Learner = new LearningSystem();
        Learner.setup(state, p_learn);
        
        // Auto fitness?
        Parameter p_auto = base.push("autofitness");
        if (state.parameters.exists(p_auto, null))
        {
        	AutoFitness = (IEASSimpleAutomaticFitness)
        		(state.parameters.getInstanceForParameterEq(p_auto,null,IEASSimpleAutomaticFitness.class));
        	AutoFitness.setup(state, p_auto);
        }
        
        // Not much else to check or setup.
        state.output.systemMessage("Problem initialized.");
    }
	
	// This is done across multiple threads.
	// It's been relegated to produce the image phenotype.
	// User evaluation is managed by the IEASEvaluator, which calls our userEvaluate() method.
	@Override
	public void evaluate(EvolutionState state, Individual ind, int subpopulation, int threadnum) 
	{
		if(ind.evaluated) return;
		
		if(!( ind instanceof IEASAnalyzedIndividual))
			state.output.fatal("GPIndividual class must subclass from " + IEASAnalyzedIndividual.class);
			
		// Genotype --> Phenotype
		IEASAnalyzedIndividual indv = (IEASAnalyzedIndividual)ind;
		indv.Rendering = RenderImage(state, indv, subpopulation, threadnum, Default_Image_Size);
		
		// Analysis!
		indv.Analyze();
		
		// Default fitness.
        KozaFitness f = ((KozaFitness)ind.fitness);
        f.hits = 0;
        f.setStandardizedFitness(state, 1.0); // 0.0==>Good, 1.0==>Bad        
        
        // If we're using a fixed fitness scheme (for Learning System testing)
        // Suppress user interaction and assign fitness manually
        if(AutoFitness != null)
        	AutoFitness.SetFitness(state, ind);
        
        // Mark as evaluated.
        ind.evaluated = true;
	}
	
	
	protected static final int Default_Image_Size = 128;
	
	// We expect a [0.0,1.0] range. Make it [0,255] (fairly). 
	private int FitToChannelRange(double v)
	{
		v = Math.min(Math.max( 0.0, v), 1.0 );
		return (int)Math.ceil( v * 255.0 );
	}
	
	// Separated, and with parameterized size, to allow for calls from the window
	// (ie. user wants to look at image details, or save a huge image)
	public BufferedImage RenderImage(EvolutionState state, IEASAnalyzedIndividual ind, int subpop, int threadnum, int size)
	{
		final int Image_Size_X = size;
		final int Image_Size_Y = size;
		// Percentage of image that a pixel takes up
		final double Image_Step_X = 2.0 / ((double) Image_Size_X);
		final double Image_Step_Y = 2.0 / ((double) Image_Size_Y);
		
		MultiData input = (MultiData)(this.input);
		BufferedImage outImage = new BufferedImage(Image_Size_X, Image_Size_Y, BufferedImage.TYPE_INT_RGB);
		
		for(int y = 0; y < Image_Size_Y; y++)
		for(int x = 0; x < Image_Size_X; x++)
		{
			int r, g, b, rgb = 0;
			Current_Pos_X = -1.0 + Image_Step_X * x;
			Current_Pos_Y = -1.0 + Image_Step_Y * y;
			
			if( ind.trees.length <3 )
			{	
				// Oh ... uh, okay then? I guess grayscale?
				ind.trees[0].child.eval(state,threadnum,input,stack,ind,this);
				r = g = b = FitToChannelRange(input.GetD());
			}
			else
			{
				ind.trees[0].child.eval(state,threadnum,input,stack,ind,this);
				r = FitToChannelRange(input.GetD());
				ind.trees[1].child.eval(state,threadnum,input,stack,ind,this);
				g = FitToChannelRange(input.GetD());
				ind.trees[2].child.eval(state,threadnum,input,stack,ind,this);
				b = FitToChannelRange(input.GetD());
			}
			
			rgb = (r<<16) + (g<<8) + (b);
			outImage.setRGB(x, y, rgb);
		}
		
		return outImage;
	}
	
	@Override
	public void prepareToUserEvaluate(final EvolutionState state)
	{	
		if(	AutoFitness != null)
			if (state.generation != state.numGenerations-1 || !AutoFitness.Show_Final_Gen)
				return;
		
		IEASLearningWindow.Instance().BindEvolutionState(state);
		IEASLearningWindow.Instance().setVisible(true);
	}
	@Override
	public void userEvaluate(final EvolutionState state)
	{
		if(AutoFitness != null)
		{
			IEASLearningWindow.setUserDeferred( AutoFitness.IsDeferal() );
			AutoFitness.NextGeneration();
			
			if (state.generation != state.numGenerations-1 || !AutoFitness.Show_Final_Gen)
				return;
		}
		
		IEASLearningWindow.Instance().RefreshIndividuals();
		IEASLearningWindow.Instance().WaitForUser();
	}
	@Override
	public void userEvaluationFinished(final EvolutionState state)
	{
		if(isUserDeferringRating()) return;
		
		List<Instance> instance_collection = new ArrayList<Instance>();
		for( ec.Subpopulation subpop : state.population.subpops )
		for( ec.Individual ind : subpop.individuals )
		{
			IEASAnalyzedIndividual indv = (IEASAnalyzedIndividual)ind;
			KozaFitness f = ((KozaFitness)ind.fitness);
			Instance inst = Learner.IndividualToInstance(indv, f.hits);
			instance_collection.add(inst);
		}
		Instances insts = new Instances("", Learner.GetFullAttributeList(), instance_collection.size());
		for(Instance inst : instance_collection)
			insts.add(inst);
		
		Learner.Train(insts);
	}
	public void learnerEvaluate(final EvolutionState state)
	{
		if(!isUserDeferringRating()) return;
		
		List<Instance> instance_collection = new ArrayList<Instance>();
		int idx = 0;
		for( ec.Subpopulation subpop : state.population.subpops )
		for( ec.Individual ind : subpop.individuals )
		{
			IEASAnalyzedIndividual indv = (IEASAnalyzedIndividual)ind;
			Instance inst = Learner.IndividualToInstance(indv);
			instance_collection.add(idx,inst);
			idx++;
		}
		
		// Add instances to an Instances collection for the normalization to work
		Instances insts = new Instances("", Learner.GetFullAttributeList(), instance_collection.size());
		for(int i = 0; i < idx; i++)
			insts.add(i,instance_collection.get(i));		
		// Try to normalize
		insts = Learner.Normalize(insts);
		
		idx = 0;
		for( ec.Subpopulation subpop : state.population.subpops )
		for( ec.Individual ind : subpop.individuals )
		{
			Instance inst = insts.get(idx);
			Learner.Classify(inst);
			int hits = Learner.GetFitness(inst);
			double fitness = ((double)(hits-1)) / 4.0;
			KozaFitness f = ((KozaFitness)ind.fitness);
	        f.hits = hits;
	        f.setStandardizedFitness(state, 1.0 - fitness);
	        idx++;
		}
	}
	public boolean isUserDeferringRating() 
	{
		return IEASLearningWindow.hasUserDeferred();
	}
	@Override
	public boolean userExited()
	{
		return IEASLearningWindow.hasUserExited();
	}
	
	
}
