// COSC 4P76 - Project
// Michael Gircys (mg12vp, 5269550)
// 2016-04-21

package ec.app.COSC4P76_Proj;

import java.util.ArrayList;
import java.util.List;

import ec.EvolutionState;
import ec.util.Parameter;
import weka.core.Attribute;
import weka.core.DenseInstance;
import weka.core.Instance;
import weka.core.Instances;
import weka.attributeSelection.InfoGainAttributeEval;
import weka.attributeSelection.Ranker;
import weka.classifiers.evaluation.Evaluation;
import weka.classifiers.functions.MultilayerPerceptron;
import weka.classifiers.trees.J48;
import weka.filters.Filter;
import weka.filters.supervised.attribute.AttributeSelection;
import weka.filters.unsupervised.attribute.Normalize;


// A higher level classification system employing a combination
// of Decision trees and ANNs.
public class LearningSystem 
{

	// Reference to the class attribute;
	// PCA / feature reduction system objects 
	public Attribute 				ClassAttribute;	
	public InfoGainAttributeEval	AttributeEval;
	public Ranker					AttributeRanker;
	public AttributeSelection		AttributeFilter;
	
	// Individual classifiers
	public J48						Learner_Tree 	= null;
	public MultilayerPerceptron		Learner_ANN  	= null;
	
	// Classifier evaluations (performance measures)
	public Evaluation				Evaluation_Tree	= null;
	public Evaluation				Evaluation_ANN 	= null;
	
	// Normalization options
	public Normalize            	Filter_Normalize;
	public boolean              	isNormalizing;
	
	// Training sample reuse options
	public boolean                  isDatasetCumulative;
	public Instances				Dataset_Cumulative = null;
	
	
	// Learning system initialization
	// Called during problem initialization
	/* Params (from ecj .params file):
	 * 		do_normalize
	 * 		keepolddataset
	 * 		pca_features
	 * 		J48 Tree Options:
	 * 			tree.confidence
	 * 			tree.minnumobj
	 * 			tree.numfolds
	 * 			tree.reducederrorpruning
	 * 			tree.saveinstancedata
	 * 			treeseed
	 * 		ANN Options:
	 * 			hiddenlayers
	 * 			learningrate
	 * 			momentumrate
	 * 			epochs
	*/
	public void setup(final EvolutionState state, final Parameter base)
	{
		int attrCount = (FeatureRegionType.values().length * FeatureType.values().length) + 1;
		
		// Specify class attribute
		List<String> classifications = new ArrayList<String>();
		for(int i = 1; i <= 5; i++)
			classifications.add(String.valueOf(i));
		ClassAttribute = new Attribute( "Hits", classifications, attrCount-1 );
		
		// Normalize data
		isNormalizing = state.parameters.getBoolean(base.push("do_normalize"), null, true);
		
		// Reuse previous training samples
		isDatasetCumulative = state.parameters.getBoolean(base.push("keepolddataset"), null, false);
		
		// // Initialize and configure PCA options if required
		int PCA_Features = state.parameters.getIntWithDefault(base.push("pca_features"), null, 0);
		if(PCA_Features > 0)
		{
			//<https://weka.wikispaces.com/Use+Weka+in+your+Java+code#Attribute%20selection>
			AttributeEval = new InfoGainAttributeEval();
			AttributeRanker = new Ranker();
			AttributeRanker.setNumToSelect(PCA_Features);
			AttributeFilter = new AttributeSelection();
			AttributeFilter.setEvaluator(AttributeEval);
			AttributeFilter.setSearch(AttributeRanker);
		}
		
		// Initialize and configure Tree options
		Parameter p_tree = base.push("tree");
		double  tree_confidence          = state.parameters.getDoubleWithDefault(p_tree.push("confidence"), null, 0.25);
		int     tree_minnumobj           = state.parameters.getIntWithDefault(p_tree.push("minnumobj"),     null, 2);
		int     tree_numfolds            = state.parameters.getIntWithDefault(p_tree.push("numfolds"),      null, 3);
		boolean tree_reducederrorpruning = state.parameters.getBoolean(p_tree.push("reducederrorpruning"),  null, false);
		boolean tree_saveinstancedata    = state.parameters.getBoolean(p_tree.push("saveinstancedata"),     null, false);
		boolean tree_subtreeraising      = state.parameters.getBoolean(p_tree.push("subtreeraising"),       null, true);
		int     tree_seed                = state.parameters.getIntWithDefault(p_tree.push("seed"),          null, 1);
		Learner_Tree = new J48();
        Learner_Tree.setConfidenceFactor((float)tree_confidence);
        Learner_Tree.setMinNumObj(tree_minnumobj);
        Learner_Tree.setNumFolds(tree_numfolds);
        Learner_Tree.setReducedErrorPruning(tree_reducederrorpruning);
        Learner_Tree.setSaveInstanceData(tree_saveinstancedata);
        Learner_Tree.setSubtreeRaising(tree_subtreeraising);
        Learner_Tree.setSeed(tree_seed);
        
        // Initialize and configure ANN options
        Parameter p_ann = base.push("ann");
        String ann_hiddenlayers = state.parameters.getStringWithDefault(p_ann.push("hiddenlayers"), null, "10");
        double ann_learningrate = state.parameters.getDoubleWithDefault(p_ann.push("learningrate"), null, 0.3);
        double ann_momentumrate = state.parameters.getDoubleWithDefault(p_ann.push("momentumrate"), null, 0.2);
        int    ann_epochs       = state.parameters.getIntWithDefault(p_ann.push("epochs"),          null, 500);
        Learner_ANN = new MultilayerPerceptron();
        Learner_ANN.setHiddenLayers(ann_hiddenlayers);
        Learner_ANN.setLearningRate(ann_learningrate);
        Learner_ANN.setMomentum(ann_momentumrate);
        Learner_ANN.setTrainingTime(ann_epochs);
        Learner_ANN.setAutoBuild(true);
	}
	
	// Create and Instance (sample) from IEASAnalyzedIndividual features and fitness
	public Instance IndividualToInstance(IEASAnalyzedIndividual ind)
	{
		return IndividualToInstance(ind, 0);
	}
	public Instance IndividualToInstance(IEASAnalyzedIndividual ind, int fitClass)
	{
		int attrCount = (FeatureRegionType.values().length * FeatureType.values().length) + 1;
		Instance inst = new DenseInstance(attrCount);
		
		int idx = 0;
		for(FeatureRegionType region : FeatureRegionType.values())
		for(FeatureType feature : FeatureType.values())
		{
			double val = ind.GetFeature(region, feature);
			if (Double.isNaN(val) || Double.isInfinite(val)) val = 0.0;
			inst.setValue(idx, val);
			idx++;
		}
		if(fitClass > 0)
		{
			int fitClassInd = ClassAttribute.indexOfValue(String.valueOf(fitClass));
			inst.setValue(idx, fitClassInd);
		}
		
		return inst;
	}
	
	// Normalize a set of instances to [0,1]
	public Instances Normalize(Instances insts)
	{
		Instances insts_filtered = insts;
		
		if(!isNormalizing)
			return insts_filtered;
		
		try {
			if(Filter_Normalize==null)
			{
				Filter_Normalize = new Normalize();
				Filter_Normalize.setInputFormat(insts);
			}
			insts_filtered = Filter.useFilter(insts, Filter_Normalize);
		} catch (Exception e1) 
		{ 
			e1.printStackTrace(); 
		}
		
		return insts_filtered;
	}
	
	// Use feature reduction methods to reduce dimensionality 
	public Instances ReduceDimension(Instances insts)
	{
		Instances insts_filtered = insts;
		
		if(AttributeFilter != null)
		{
			try {
				AttributeFilter.setInputFormat(insts);
				insts_filtered = Filter.useFilter(insts, AttributeFilter);
			} catch (Exception e) {
				e.printStackTrace();
			}
		}
		
		return insts_filtered;
	}
	
	// Classify a single instance using the best individual classifier
	public void Classify(Instance inst)
	{ 
		boolean useLearner_ANN = Evaluation_ANN.correct() >= Evaluation_Tree.correct();
		
		Instance inst_filtered = inst;
		// Already Normalized as a set in IEASLerning.learnerEvaluate().
		
		double res = 0;
		try {
			if(useLearner_ANN)
					res = Learner_ANN.classifyInstance(inst_filtered);
			else 	res = Learner_Tree.classifyInstance(inst_filtered);
		} catch (Exception e) {
			e.printStackTrace();
		}
		
		inst.setValue( ClassAttribute.index(), (int)Math.round(res) );
	}
	
	// Create new classifier instances incorporating a new dataset
	public void Train(Instances insts)
	{
		// Ensure class attribute has been set
		insts.setClassIndex(ClassAttribute.index());
		
		// Are we including stored instance data from previous generations?
		if(isDatasetCumulative)
		{
			if(Dataset_Cumulative == null)
					Dataset_Cumulative = Instances.mergeInstances(Dataset_Cumulative,insts);
			else	Dataset_Cumulative = insts;
			insts = Dataset_Cumulative;
		}
		
		// Normalize data
		insts = Normalize(insts);
		
		// Try PCA?
		insts = ReduceDimension(insts);
		
		// Update and evaluate each classifier
		try {
			Learner_ANN.buildClassifier(insts);
			Evaluation_ANN = new Evaluation(insts);
			Evaluation_ANN.evaluateModel(Learner_ANN, insts);
			System.out.println(Evaluation_ANN.toSummaryString());
		} catch (Exception e) {
			e.printStackTrace();
		}
		try {
			Learner_Tree.buildClassifier(insts);
			Evaluation_Tree = new Evaluation(insts);
			Evaluation_Tree.evaluateModel(Learner_Tree, insts);
			System.out.println(Evaluation_Tree.toSummaryString());
		} catch (Exception e) {
			e.printStackTrace();
		}
		
	}
	
	// Retrieve fitness from an individual's classification 
	public int GetFitness(Instance inst)
	{
		int hitsIndex = (int)Math.round(inst.classValue());
		int hits = Integer.parseInt( ClassAttribute.value(hitsIndex) );
		return hits;
	}
	
	// Produce a full list of attributes from the list of calculated features
	public ArrayList<Attribute> GetFullAttributeList()
	{
		ArrayList<Attribute> attrList = new ArrayList<Attribute>();
		
		int idx = 0;
		for(FeatureRegionType region : FeatureRegionType.values())
		for(FeatureType feature : FeatureType.values())
		{
			String featureName = region.toString()+"_"+feature.toString();
			Attribute attr = new Attribute(featureName, idx);
			attrList.add(attr);
			idx++;
		}
		attrList.add(ClassAttribute);
		
		return attrList;
	}
	
}
